"""Pytest test config tools."""

import logging
import os
from pathlib import Path

import pytest

from disease.database import AWS_ENV_VAR_NAME, create_db
from disease.database.database import AbstractDatabase
from disease.etl.base import Base
from disease.query import QueryHandler
from disease.schemas import Disease, MatchType, SourceSearchMatches

_logger = logging.getLogger(__name__)


def pytest_collection_modifyitems(items):
    """Modify test items in place to ensure test modules run in a given order.
    When creating new test modules, be sure to add them here.
    """
    module_order = [
        "test_schemas",
        "test_mondo",
        "test_do",
        "test_ncit",
        "test_omim",
        "test_oncotree",
        "test_merge",
        "test_database",
        "test_query",
        "test_endpoints",
        "test_emit_warnings",
    ]
    # remember to add new test modules to the order constant:
    assert len(module_order) == len(list(Path(__file__).parent.rglob("test_*.py")))
    items.sort(key=lambda i: module_order.index(i.module.__name__))


def pytest_addoption(parser):
    """Add custom commands to pytest invocation.
    See https://docs.pytest.org/en/7.1.x/reference/reference.html#parser
    """
    parser.addoption(
        "--verbose-logs",
        action="store_true",
        default=False,
        help="show noisy module logs",
    )


def pytest_configure(config):
    """Configure pytest setup."""
    if not config.getoption("--verbose-logs"):
        logging.getLogger("botocore").setLevel(logging.ERROR)
        logging.getLogger("boto3").setLevel(logging.ERROR)
        logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)


TEST_ROOT = Path(__file__).resolve().parents[1]
TEST_DATA_DIRECTORY = TEST_ROOT / "tests" / "data"
IS_TEST_ENV = os.environ.get("DISEASE_TEST", "").lower() == "true"


def pytest_sessionstart():
    """Wipe DB before testing if in test environment."""
    if IS_TEST_ENV:
        if os.environ.get(AWS_ENV_VAR_NAME):
            msg = f"Cannot have both DISEASE_TEST and {AWS_ENV_VAR_NAME} set."
            raise AssertionError(msg)
        db = create_db()
        db.drop_db()
        db.initialize_db()


@pytest.fixture(scope="session")
def is_test_env():
    """If true, currently in test environment (i.e. okay to overwrite DB). Downstream
    users should also make sure to check if in a production environment.

    Provided here to be accessible directly within test modules.
    """
    return IS_TEST_ENV


@pytest.fixture(scope="module")
def database():
    """Provide a database instance to be used by tests."""
    db = create_db()
    yield db
    db.close_connection()


@pytest.fixture(scope="module")
def test_source(database: AbstractDatabase, is_test_env: bool):
    """Provide query endpoint for testing sources. If DISEASE_TEST is set, will try to
    load DB from test data.

    :param database: test database instance
    :param is_test_env: if true, load from test data
    :return: factory function that takes an ETL class instance and returns a query
    endpoint.
    """

    def test_source_factory(EtlClass: Base):  # noqa: N803
        if IS_TEST_ENV:
            _logger.debug("Reloading DB with data from %s", TEST_DATA_DIRECTORY)
            test_class = EtlClass(
                database, TEST_DATA_DIRECTORY / EtlClass.__name__.lower()
            )
            test_class.perform_etl(use_existing=True)

        class QueryGetter:
            def __init__(self):
                self.query_handler = QueryHandler(database)
                self._src_name = EtlClass.__name__

            def search(self, query_str: str):
                resp = self.query_handler.search(query_str, incl=self._src_name)
                return resp.source_matches[self._src_name]

        return QueryGetter()

    return test_source_factory


def _compare_records(actual: Disease, fixt: Disease):
    """Check that identity records are identical."""
    assert actual.concept_id == fixt.concept_id
    assert actual.label == fixt.label

    assert (actual.aliases is None) == (fixt.aliases is None)
    if (actual.aliases is not None) and (fixt.aliases is not None):
        assert set(actual.aliases) == set(fixt.aliases)

    assert (actual.xrefs is None) == (fixt.xrefs is None)
    if (actual.xrefs is not None) and (fixt.xrefs is not None):
        assert set(actual.xrefs) == set(fixt.xrefs)

    assert (actual.associated_with is None) == (fixt.associated_with is None)
    if (actual.associated_with is not None) and (fixt.associated_with is not None):
        assert set(actual.associated_with) == set(fixt.associated_with)

    assert (actual.pediatric_disease is None) == (fixt.pediatric_disease is None)
    if (actual.pediatric_disease is not None) and (fixt.pediatric_disease is not None):
        assert actual.pediatric_disease == fixt.pediatric_disease

    assert (actual.oncologic_disease is None) == (fixt.oncologic_disease is None)
    if (actual.oncologic_disease is not None) and (fixt.oncologic_disease is not None):
        assert actual.oncologic_disease == fixt.oncologic_disease


@pytest.fixture(scope="session")
def compare_records():
    """Provide record comparison function"""
    return _compare_records


def _compare_response(
    response: SourceSearchMatches,
    match_type: MatchType,
    fixture: Disease | None = None,
    fixture_list: list[Disease] | None = None,
    num_records: int = 0,
):
    """Check that test response is correct. Only 1 of {fixture, fixture_list}
    should be passed as arguments. num_records should only be passed with fixture_list.

    :param Dict response: response object returned by QueryHandler
    :param MatchType match_type: expected match type
    :param Disease fixture: single Disease object to match response against
    :param List[Disease] fixture_list: multiple Disease objects to match response
        against
    :param int num_records: expected number of records in response. If not given, tests
        for number of fixture Diseases given (ie, 1 for single fixture and length of
        fixture_list otherwise)
    """
    if fixture and fixture_list:
        msg = "Args provided for both `fixture` and `fixture_list`"
        raise Exception(msg)
    if not fixture and not fixture_list:
        msg = "Must pass 1 of {fixture, fixture_list}"
        raise Exception(msg)
    if fixture and num_records:
        msg = "`num_records` should only be given with `fixture_list`."
        raise Exception(msg)

    assert response.match_type == match_type
    if fixture:
        assert len(response.records) == 1
        _compare_records(response.records[0], fixture)
    elif fixture_list:
        if not num_records:
            assert len(response.records) == len(fixture_list)
        else:
            assert len(response.records) == num_records
        for fixt in fixture_list:
            for record in response.records:
                if fixt.concept_id == record.concept_id:
                    _compare_records(record, fixt)
                    break
            raise AssertionError  # test fixture not found in response


@pytest.fixture(scope="session")
def compare_response():
    """Provide response comparison function"""
    return _compare_response



"""Provides a CLI util to make updates to normalizer database."""

import logging
import os
from collections.abc import Collection
from pathlib import Path
from timeit import default_timer as timer

import click

from disease import SOURCES_FOR_MERGE, SOURCES_LOWER_LOOKUP
from disease.database.database import (
    AbstractDatabase,
    DatabaseException,
    DatabaseReadException,
    DatabaseWriteException,
    create_db,
)
from disease.schemas import SourceName

_logger = logging.getLogger(__name__)


def _configure_logging() -> None:
    """Configure logging."""
    logging.basicConfig(
        filename=f"{__package__}.log",
        format="[%(asctime)s] - %(name)s - %(levelname)s : %(message)s",
    )
    logging.getLogger(__package__).setLevel(logging.DEBUG)


@click.command()
@click.option("--db_url", help="URL endpoint for the application database.")
@click.option("--verbose", "-v", is_flag=True, help="Print result to console if set.")
def check_db(db_url: str, verbose: bool = False) -> None:
    """Perform basic checks on DB health and population. Exits with status code 1
    if DB schema is uninitialized or if critical tables appear to be empty.

    \f
    :param db_url: URL to normalizer database
    :param verbose: if true, print result to console
    """  # noqa: D301
    _configure_logging()
    db = create_db(db_url, False)
    if not db.check_schema_initialized():
        if verbose:
            click.echo("Health check failed: DB schema uninitialized.")
        click.get_current_context().exit(1)

    if not db.check_tables_populated():
        if verbose:
            click.echo("Health check failed: DB is incompletely populated.")
        click.get_current_context().exit(1)

    msg = "DB health check successful: tables appear complete."
    if verbose:
        click.echo(msg)
    _logger.info(msg)


@click.command()
@click.option("--data_url", help="URL to data dump")
@click.option("--db_url", help="URL endpoint for the application database.")
def update_from_remote(data_url: str | None, db_url: str) -> None:
    """Update data from remotely-hosted DB dump. By default, fetches from latest
    available dump on VICC S3 bucket; specific URLs can be provided instead by
    command line option or DISEASE_NORM_REMOTE_DB_URL environment variable.

    \f
    :param data_url: user-specified location to pull DB dump from
    :param db_url: URL to normalizer database
    """  # noqa: D301
    _configure_logging()
    if not click.confirm("Are you sure you want to overwrite existing data?"):
        click.get_current_context().exit()
    if not data_url:
        data_url = os.environ.get("DISEASE_NORM_REMOTE_DB_URL")
    db = create_db(db_url, False)
    try:
        db.load_from_remote(data_url)
    except NotImplementedError:
        click.echo(
            "Error: Fetching remote data dump not supported for "
            f"{db.__class__.__name__}"
        )
        click.get_current_context().exit(1)
    except DatabaseException as e:
        click.echo(f"Encountered exception during update: {e!s}")
        click.get_current_context().exit(1)
    _logger.info("Successfully loaded data from remote snapshot.")


@click.command()
@click.option(
    "--output_directory",
    "-o",
    help="Output location to write to",
    type=click.Path(exists=True, path_type=Path),
)
@click.option("--db_url", help="URL endpoint for the application database.")
def dump_database(output_directory: Path, db_url: str) -> None:
    """Dump data from database into file.

    \f
    :param output_directory: path to existing directory
    :param db_url: URL to normalizer database
    """  # noqa: D301
    _configure_logging()
    if not output_directory:
        output_directory = Path()

    db = create_db(db_url, False)
    try:
        db.export_db(output_directory)
    except NotImplementedError:
        click.echo(
            f"Error: Dumping data to file not supported for {db.__class__.__name__}"
        )
        click.get_current_context().exit(1)
    except DatabaseException as e:
        click.echo(f"Encountered exception during update: {e!s}")
        click.get_current_context().exit(1)
    _logger.info("Database dump successful.")


def _update_sources(
    sources: Collection[SourceName],
    db: AbstractDatabase,
    update_merged: bool,
    from_local: bool,
) -> None:
    """Update selected normalizer sources.

    :param sources: names of sources to update
    :param db: database instance
    :param update_merged: if true, retain processed records to use in updating merged
        records
    :param from_local: if true, use locally available data only
    """
    processed_ids = []
    for n in sources:
        delete_time = _delete_source(n, db)
        _load_source(n, db, delete_time, processed_ids, from_local)

    if update_merged:
        _load_merge(db, set(processed_ids))


def _delete_source(n: SourceName, db: AbstractDatabase) -> float:
    """Delete individual source data.

    :param n: name of source to delete
    :param db: database instance
    :return: time taken (in seconds) to delete
    """
    msg = f"Deleting {n.value}..."
    click.echo(f"\n{msg}")
    _logger.info(msg)
    start_delete = timer()
    db.delete_source(n)
    end_delete = timer()
    delete_time = end_delete - start_delete
    msg = f"Deleted {n.value} in {delete_time:.5f} seconds."
    click.echo(f"{msg}\n")
    _logger.info(msg)
    return delete_time


def _load_source(
    n: SourceName,
    db: AbstractDatabase,
    delete_time: float,
    processed_ids: list[str],
    from_local: bool,
) -> None:
    """Load individual source data.

    :param n: name of source
    :param db: database instance
    :param delete_time: time taken (in seconds) to run deletion
    :param processed_ids: in-progress list of processed disease IDs
    :param from_local: if true, use locally available data
    """
    msg = f"Loading {n.value}..."
    click.echo(msg)
    _logger.info(msg)
    start_load = timer()

    # used to get source class name from string
    try:
        from disease.etl import DO, OMIM, Mondo, NCIt, OncoTree  # noqa: F401
    except ModuleNotFoundError as e:
        click.echo(
            f"Encountered ModuleNotFoundError attempting to import {e.name}. Are ETL dependencies installed?"
        )
        click.get_current_context().exit()
    SourceClass = eval(n.value)  # noqa: N806 S307 PGH001

    source = SourceClass(database=db, silent=False)
    processed_ids += source.perform_etl(use_existing=from_local)
    end_load = timer()
    load_time = end_load - start_load
    msg = f"Loaded {n.value} in {load_time:.5f} seconds."
    click.echo(msg)
    _logger.info(msg)
    msg = f"Total time for {n.value}: {(delete_time + load_time):.5f} seconds."
    click.echo(msg)
    _logger.info(msg)


def _delete_normalized_data(database: AbstractDatabase) -> None:
    """Delete normalized concepts

    :param database: DB instance
    """
    click.echo("\nDeleting normalized records...")
    start_delete = timer()
    try:
        database.delete_normalized_concepts()
    except (DatabaseReadException, DatabaseWriteException) as e:
        click.echo(f"Encountered exception during normalized data deletion: {e}")
        click.get_current_context().exit(1)
    end_delete = timer()
    delete_time = end_delete - start_delete
    click.echo(f"Deleted normalized records in {delete_time:.5f} seconds.")


def _load_merge(db: AbstractDatabase, processed_ids: set[str]) -> None:
    """Load merged concepts

    :param db: database instance
    :param processed_ids: in-progress list of processed disease IDs
    """
    start = timer()
    _delete_normalized_data(db)
    if not processed_ids:
        processed_ids = set()
        for source in SOURCES_FOR_MERGE:
            processed_ids |= db.get_all_concept_ids(source)

    try:
        from disease.etl.merge import Merge
    except ModuleNotFoundError as e:
        click.echo(
            f"Encountered ModuleNotFoundError attempting to import {e.name}. Are ETL dependencies installed?"
        )
        click.get_current_context().exit()

    merge = Merge(database=db)
    click.echo("Constructing normalized records...")
    merge.create_merged_concepts(processed_ids)
    end = timer()
    click.echo(
        f"Merged concept generation completed in " f"{(end - start):.5f} seconds"
    )


@click.command()
@click.option("--sources", help="The source(s) you wish to update separated by spaces.")
@click.option("--aws_instance", is_flag=True, help="Using AWS DynamodDB instance.")
@click.option("--db_url", help="URL endpoint for the application database.")
@click.option("--update_all", is_flag=True, help="Update all normalizer sources.")
@click.option(
    "--update_merged",
    is_flag=True,
    help="Update concepts for normalize endpoint from accepted sources.",
)
@click.option(
    "--from_local",
    is_flag=True,
    default=False,
    help="Use most recent local source data instead of fetching latest versions.",
)
def update_db(
    sources: str,
    aws_instance: bool,
    db_url: str,
    update_all: bool,
    update_merged: bool,
    from_local: bool,
) -> None:
    """Update selected normalizer source(s) in the disease database.

    \f
    :param sources: names of sources to update, comma-separated
    :param aws_instance: if true, use cloud instance
    :param db_url: URI pointing to database
    :param update_all: if true, update all sources (ignore `normalizer` parameter)
    :param update_merged: if true, update normalized records
    :param from_local: if true, use locally available data only
    """  # noqa: D301
    _configure_logging()
    db = create_db(db_url, aws_instance)

    if update_all:
        _update_sources(list(SourceName), db, update_merged, from_local)
    elif not sources:
        if update_merged:
            _load_merge(db, set())
        else:
            ctx = click.get_current_context()
            click.echo(
                "Must either enter 1 or more sources, or use `--update_all` parameter"
            )
            click.echo(ctx.get_help())
            ctx.exit()
    else:
        sources_split = sources.lower().split()

        if len(sources_split) == 0:
            msg = "Must enter one or more source names"
            raise Exception(msg)

        non_sources = set(sources_split) - set(SOURCES_LOWER_LOOKUP)

        if len(non_sources) != 0:
            msg = f"Not valid source(s): {non_sources}"
            raise Exception(msg)

        sources_to_update = {SourceName(SOURCES_LOWER_LOOKUP[s]) for s in sources_split}
        _update_sources(sources_to_update, db, update_merged, from_local)
    _logger.info("Database update successful.")


if __name__ == "__main__":
    update_db()



"""Perform basic tests of endpoint branches.
We already have tests and data validation to ensure correctness of the underlying
response objects -- here, we're checking for bad branch logic and for basic assurances
that routes integrate correctly with query methods.
"""

import pytest
from fastapi.testclient import TestClient

from disease.main import app


@pytest.fixture(scope="module")
def api_client():
    """Provide test client fixture."""
    return TestClient(app)


def test_search(api_client):
    """Test /search endpoint."""
    response = api_client.get("/disease/search?q=neuroblastoma")
    assert response.status_code == 200
    assert (
        response.json()["source_matches"]["Mondo"]["records"][0]["concept_id"]
        == "mondo:0005072"
    )

    response = api_client.get("/disease/search?q=neuroblastoma&incl=sdkl")
    assert response.status_code == 422

    response = api_client.get("/disease/search")
    assert response.status_code == 422


def test_normalize(api_client):
    """Test /normalize endpoint."""
    response = api_client.get("/disease/normalize?q=neuroblastoma")
    assert response.status_code == 200
    assert response.json()["disease"]["primaryCode"] == "ncit:C3270"

    response = api_client.get("/disease/normalize")
    assert response.status_code == 422



"""Main application for FastAPI"""

import html

from fastapi import FastAPI, HTTPException, Query
from fastapi.openapi.utils import get_openapi

from disease import __version__
from disease.database.database import create_db
from disease.query import InvalidParameterException, QueryHandler
from disease.schemas import NormalizationService, SearchService

db = create_db()
query_handler = QueryHandler(db)

app = FastAPI(
    docs_url="/disease",
    openapi_url="/disease/openapi.json",
    swagger_ui_parameters={"tryItOutEnabled": True},
)


def custom_openapi() -> dict | None:
    """Generate custom fields for OpenAPI response"""
    if app.openapi_schema:
        return app.openapi_schema
    openapi_schema = get_openapi(
        title="The VICC Disease Normalizer",
        version=__version__,
        openapi_version="3.0.3",
        description="Normalize disease terms.",
        routes=app.routes,
    )
    #    openapi_schema['info']['license'] = {  # TODO
    #        "name": "Name-of-license",
    #        "url": "http://www.to-be-determined.com"
    #    }
    openapi_schema["info"]["contact"] = {
        "name": "Alex H. Wagner",
        "email": "Alex.Wagner@nationwidechildrens.org",
        "url": "https://www.nationwidechildrens.org/specialties/institute-for-genomic-medicine/research-labs/wagner-lab",
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


app.openapi = custom_openapi

# endpoint description text
get_matches_summary = "Given query, provide highest matches from " "each source."
search_descr = "Search for disease term."
response_descr = "A response to a validly-formed query."
q_descr = "Disease term to search for."
incl_descr = (
    "Comma-separated list of source names to include in "
    "response. Will exclude all other sources. Will return HTTP "
    "status code 422: Unprocessable Entity if both 'incl' and "
    "'excl' parameters are given."
)
excl_descr = (
    "Comma-separated list of source names to exclude in "
    "response. Will include all other sources. Will return HTTP "
    "status code 422: Unprocessable Entity if both 'incl' and"
    "'excl' parameters are given."
)
normalize_description = (
    "Return merged strongest-match concept for query " "string provided by user."
)


@app.get(
    "/disease/search",
    summary=get_matches_summary,
    description=search_descr,
    operation_id="getQueryResponse",
    response_description=response_descr,
    response_model=SearchService,
    response_model_exclude_none=True,
)
def search(
    q: str = Query(..., description=q_descr),
    incl: str | None = Query("", description=incl_descr),
    excl: str | None = Query("", description=excl_descr),
) -> SearchService:
    """For each source, return strongest-match concepts for query string
    provided by user.

    :param q: disease search term
    :param incl: sources to include
    :param excl: sources to excl
    :return: search results
    """
    try:
        response = query_handler.search(html.unescape(q), incl=incl, excl=excl)
    except InvalidParameterException as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response


merged_matches_summary = "Given query, provide normalized record."
merged_response_descr = "A response to a validly-formed query."
merged_q_descr = "Disease to normalize."


@app.get(
    "/disease/normalize",
    summary=merged_matches_summary,
    operation_id="getQuerymergedResponse",
    response_description=merged_response_descr,
    response_model=NormalizationService,
    description=normalize_description,
    response_model_exclude_none=True,
)
def normalize(q: str = Query(..., description=merged_q_descr)) -> NormalizationService:
    """Return strongest-match normalized concept for query string provided by
    user.
    :param q: disease search term
    """
    try:
        response = query_handler.normalize(q)
    except InvalidParameterException as e:
        raise HTTPException(status_code=422, detail=str(e)) from None
    return response
